
(** Libraries *)
Require Import ssreflect.
Require Import LocConfTac.


(** Proof of local confluence of the algebraic fragment of the rewrite
    system Llin *)


Section Llin.

(** The scalar system *)

Variable scalar : Set.

Variable Sadd : scalar -> scalar -> scalar.
Variable Smul : scalar -> scalar -> scalar.
Variable S0 : scalar.
Variable S1 : scalar. 

Notation "A + B" := (Sadd A B) : scalar_scope.
Notation "A * B" := (Smul A B) : scalar_scope.

Open Scope scalar_scope.

Hypothesis S_0_1_dec : ~ S1 = S0.
Hypothesis S_0_lunit : forall a, S0 + a = a.
Hypothesis S_0_lelim : forall a, S0 * a = S0.
Hypothesis S_1_lunit : forall a, S1 * a = a.
Hypothesis S_rdistrib : forall a b c, a*(b+c) = (a*b)+(a*c).
Hypothesis S_ldistrib : forall a b c, (a+b)*c = (a*c)+(b*c).
Hypothesis S_add_assoc : forall a b c, (a+b)+c = a+(b+c).
Hypothesis S_mul_assoc : forall a b c, (a*b)*c = a*(b*c).
Hypothesis S_add_commut : forall a b, a+b = b+a.
Hypothesis S_mul_commut : forall a b, a*b = b*a.


Close Scope scalar_scope.


(** Definition of terms and values *)

Inductive term : Set :=
| T0 : term
| Tadd : term -> term -> term
| Tmul : scalar -> term -> term
| Tvar : nat -> term
| Tlambda : term -> term
| Tapply : term -> term -> term.


Notation "A +s B" := (Sadd A B) (at level 50) : term_scope.
Notation "A *s B" := (Smul A B) (at level 40) : term_scope.

Notation "A + B" := (Tadd A B) : term_scope.
Notation "A '**' B" := (Tmul A B) (at level 35) : term_scope.
Notation "@ A"   := (Tvar A) (at level 10) : term_scope.
Notation "A ; B" := (Tapply A B) (at level 30) : term_scope.
Notation "\ A"   := (Tlambda A) (at level 40) : term_scope.

Open Scope term_scope.

Inductive is_value : term -> Prop :=
| valT0 : is_value T0
| valTlambda : forall s, is_value (Tlambda s)
| valTvar : forall n, is_value (Tvar n)
| valTmulbase : forall a s, is_value s -> is_value (a ** s)
| valTadd : forall s t, is_value s -> is_value t -> is_value (s + t).

Hint Immediate valT0 valTlambda valTvar valTmulbase valTadd.


(** Try to solve a goal of the form [ |- is_value v ] when [v] is a value *) 

Ltac solve_is_value :=
  solve [
    auto
    |
    match goal with
      | [ |- is_value (?s + ?t) ] => apply valTadd; [solve_is_value | solve_is_value]
      | [ |- is_value (?a ** ?s) ] => apply valTmulbase; auto
    end
  ].

(** Try to create the hypothesis [ H : is_value s ] *)

Ltac show_is_value H s :=
  (have H : is_value s); first solve_is_value.


(* Tests

Goal forall s t:term, forall a:scalar, is_value s -> is_value t -> False.
intros s t a M N.
show_is_value H T0.
show_is_value K (T0 + (a ** t)).
show_is_value L (T0 + s).
Abort.
*)

(** Some lemmas *)

Lemma value_splitl_sum : forall s t, is_value (s + t) -> is_value s.
Proof with auto.
move => ? ? H.
inversion H...
Qed.

Lemma value_splitr_sum : forall s t, is_value (s + t) -> is_value t.
Proof  with auto.
move => ? ? H.
inversion H...
Qed.

Lemma value_split_mul : forall a s, is_value (a ** s) -> is_value s.
Proof  with auto.
move => ? ? H.
inversion H...
Qed.


(* More tests

Goal forall s t, is_value (s + t) -> False.
intros s t H.
have := (value_splitr_sum _ _ H).
have := (value_splitl_sum _ _ H).
inversion_clear H.
Abort.

Goal forall a s, is_value (a ** s) -> False.
intros a s H.
inversion_clear H.
Abort.

*)


(** Take the argument [H : is_value s] and set new arguments saying all 
   subterms are values *)

Ltac split_value H :=
  move : H;
  match goal with
    [ |- is_value (_ + _) -> _ ] => 
      move => H; 
      let K1 := fresh in 
        let K2 := fresh in (
      (have K1 := (value_splitr_sum _ _ H));
      (have K2 := (value_splitl_sum _ _ H));
      split_value K1; split_value K2
      )
    |
    [ |- is_value (_ ** _) -> _ ] => 
    move => H;
      let K1 := fresh in (
      (have K1 := (value_split_mul _ _ H));
        split_value K1
      )
    |
    _ => move => H
  end.


(* More tests

Goal forall a s t, is_value ((a ** s) + t) -> False.
intros.
split_value H.
Abort.

*)



(** The rewrite system we consider. Note that we use values. *)

Section Term.

Hypothesis R : term -> term -> Prop.

(** Elementary rules *)
Definition R_T0_runit := forall t, R (t + T0) t.
Definition R_S0_anni  := forall t, R (S0 ** t) T0.
Definition R_S1_unit  := forall t, R (S1 ** t) t.
Definition R_T0_anni  := forall a, R (a ** T0) T0.
Definition R_mul_abs  := forall a b t, R (a ** (b ** t)) ((a *s b) ** t).
Definition R_ma_dist  := forall a s t, R (a ** (s + t)) (a ** s + a ** t).
(** Factorization *)
Definition R_add_fact := forall a b t, R (a ** t + b ** t) ((a +s b) ** t).
Definition R_add_fact1 := forall a t, R (a ** t + t) ((a +s S1) ** t).
Definition R_add_fact11 := forall t, R (t + t) ((S1 +s S1) ** t).
(** Assoc. and commut. of addition *)
Definition R_add_com := forall s t, R (s + t) (t + s).
Definition R_add_rassoc := forall r s t, R ((r + s) + t) (r + (s + t)).
Definition R_add_lassoc := forall r s t, R (r + (s + t)) ((r + s) + t).
(** Congruence *)
Definition R_cong_mul := forall a s t, R s t -> R (a**s) (a**t).
Definition R_cong_ladd := forall u s t, R s t -> R (s+u) (t+u).
Definition R_cong_radd := forall u s t, R s t -> R (u+ s) (u+t).
Definition R_cong_lapp := forall u s t, R s t -> R (s;u) (t;u).
Definition R_cong_rapp := forall u s t, is_value u -> R s t -> R (u;s) (u;t).
(** Linearity of application *)
Definition R_add_app_ldist := forall r s t, is_value r -> is_value s -> is_value t -> R ((r + s);t) (r;t + s;t).
Definition R_mul_app_ldist := forall a r s, is_value r -> is_value s -> R ((a**r);s) (a**(r;s)).
Definition R_T0_app_ldist := forall s, is_value s -> R (T0;s) T0.
Definition R_add_app_rdist := forall r s t, is_value r -> is_value s -> is_value t -> R (t;(r + s)) (t;r + t;s).
Definition R_mul_app_rdist := forall a r s, is_value r -> is_value s -> R (s;(a**r)) (a**(s;r)).
Definition R_T0_app_rdist := forall s, is_value s -> R (s;T0) T0.


End Term.


Inductive R : term -> term -> Prop :=
| ax1 :R_T0_runit R
| ax2 :R_S0_anni R
| ax3 :R_S1_unit R
| ax4 :R_T0_anni R
| ax5 :R_mul_abs R
| ax6 :R_ma_dist R

| ax7 :R_add_fact R
| ax8 :R_add_fact1 R
| ax9 :R_add_fact11 R

| ax10 :R_add_com R
| ax11 :R_add_rassoc R
| ax12 :R_add_lassoc R

| ax13 :R_cong_mul R
| ax14 :R_cong_ladd R
| ax15 :R_cong_radd R
| ax16 :R_cong_lapp R
| ax17 :R_cong_rapp R

| ax18 :R_add_app_ldist R
| ax19 :R_mul_app_ldist R
| ax20 :R_T0_app_ldist R
| ax21 :R_add_app_rdist R
| ax22 :R_mul_app_rdist R
| ax23 :R_T0_app_rdist R.


Hint Immediate ax1 ax2 ax3 ax4 ax5 ax6 ax7 ax8 ax9 ax10 ax11 ax12 ax18 ax19 ax20 ax21 ax22 ax23.


(** Some properties to be able to automate the reduction process *)

Section Term_prop.

Ltac startac :=  match goal with | [ |- ?A _ ] => unfold A end; intros;
  match goal with | [ |- Rstar _ _ ?A ] => apply (Rcons _ _ _ A) end; auto.

Lemma Sax1 : R_T0_runit  (Rstar R).
Proof. by startac. Qed.

Lemma Sax2 : R_S0_anni  (Rstar R).
Proof. by startac. Qed.

Lemma Sax3 : R_S1_unit  (Rstar R).
Proof. by startac. Qed.

Lemma Sax4 : R_T0_anni  (Rstar R).
Proof. by startac. Qed.

Lemma Sax5 : R_mul_abs (Rstar R).
Proof. by startac. Qed.

Lemma Sax6 : R_ma_dist (Rstar R).
Proof. by startac. Qed.

Lemma Sax7 : R_add_fact (Rstar R).
Proof. by startac. Qed.

Lemma Sax8 : R_add_fact1 (Rstar R).
Proof. by startac. Qed.

Lemma Sax9 : R_add_fact11 (Rstar R).
Proof. by startac. Qed.

Lemma Sax10 : R_add_com (Rstar R).
Proof. by startac. Qed.

Lemma Sax11 : R_add_rassoc (Rstar R).
Proof. by startac. Qed.

Lemma Sax12 : R_add_lassoc (Rstar R).
Proof. by startac. Qed.

Lemma Sax18 : R_add_app_ldist (Rstar R).
Proof. by startac. Qed.

Lemma Sax19 : R_mul_app_ldist (Rstar R).
Proof. by startac. Qed.

Lemma Sax20 : R_T0_app_ldist (Rstar R).
Proof. by startac. Qed.

Lemma Sax21 : R_add_app_rdist (Rstar R).
Proof. by startac. Qed.

Lemma Sax22 : R_mul_app_rdist (Rstar R).
Proof. by startac. Qed.

Lemma Sax23 : R_T0_app_rdist (Rstar R).
Proof. by startac. Qed.

Lemma Sax17 : R_cong_rapp (Rstar R).
Proof.
unfold  R_cong_rapp.
intros.
elim H0.
auto.
intros.
have K : Rstar R (u; r) (u; s0).
apply R_Rstar.
apply (ax17 _ _ _ H); auto.
solve_Rstar_trans.
Qed.


End Term_prop.

Hint Immediate Sax1 Sax2 Sax3 Sax4 Sax5 Sax6 Sax7 Sax8 Sax9 Sax10 Sax11 Sax12 Sax18 Sax19 Sax20 Sax21 Sax22 Sax23 Sax17.


(** Relating values and reduction *)

Lemma R_value : forall r s, is_value r -> R r s -> is_value s.
Proof with (try solve_is_value).
elim.
(* T0 *)
move => ? ? H; inversion H.
(* + *)
move => r IHr s IHs t H K.
split_value H.
inversion K...
rewrite <- H2; auto.
rewrite -H3 in H1;  split_value H1...
rewrite -H3 in H1; split_value H1...
rewrite -H4 in H0; split_value H0...
(* ** *)
move => a s IHs t H K.
split_value H.
inversion K...
rewrite -H1...
rewrite <- H3 in H0; split_value H0...
rewrite <- H3 in H0; split_value H0...
(* @ *)
move => ? ? ? H; inversion H.
(* \ *)
move => ? ? ? ? H; inversion H.
(* ; *)
move => ? ? ? ? ? H ?; inversion H.
Qed.



Ltac auto_R_value :=
  match goal with
    | [ H : is_value ?u, K : R ?v ?u |- _ ] => move : K; auto_R_value; move => K
    | [ H : is_value ?u, K : R ?u ?v |- _ ] => 
      let L := fresh in (
        (have L : is_value v); [apply (R_value _ _ H K) | auto]; auto_R_value
      )
    | _ => idtac
  end.


(* Test

Goal forall u v w : term,
  is_value u -> R u v -> R v w -> False.
intros.
auto_R_value.
Abort.

*)


(** Tactic to prove automatically that [Rstar R a b]
    for trivial cases *)

Ltac core_step :=
  auto_R_value;
  solve [ 
      auto
    | (apply Sax18; solve_is_value)
    | (apply Sax19; solve_is_value)
    | (apply Sax20; solve_is_value)
    | (apply Sax21; solve_is_value)
    | (apply Sax22; solve_is_value)
    | (apply Sax23; solve_is_value)
    | apply (Rstar_cong _ _  (ax13 _)); core_step
    | apply (Rstar_lcong _ _ ax14); core_step
    | apply (Rstar_rcong _ _ ax15); core_step
    | apply (Rstar_lcong _ _ ax16); core_step
    | (
      let H := fresh in 
        match goal with 
          | [ |- Rstar _ (?u;_) _] => 
            idtac u;
            (have H : is_value u); [
              solve_is_value | apply (Sax17 _ _ _ H); core_step
            ]
        end
    )
    | try_Rstar_in_goal
  ].

(* Tests

Goal forall a r t,
  Rstar R ((\t);(a ** \r)) (a ** (\t);(\r)).
intros.
core_step.
Qed.

Goal forall a u v w, 
  is_value u -> is_value v -> is_value w -> 
  Rstar R (u;(a ** (v + w))) (u;(a ** v + a ** w)).
move => a u v w H K L.
core_step.
Qed.


Goal forall a r t, 
  Rstar R ((a ** \r);\t) (a ** (\r);(\t)).
intros.
core_step.
Qed.

Goal forall r s t n, 
  Rstar R ((\r + (@n + \s));\t) 
          ((\r);(\t) + (@n + \s);\t).
intros.
by core_step.
Qed.

Goal forall a r s t n, 
  Rstar R (((a ** \r) + (@n + \s));\t) 
          ((a ** \r);(\t) + (@n + \s);\t).
intros.
by core_step.
Qed.


Goal forall a, Rstar R (S0 ** a) T0.
by core_step.
Qed.




Goal forall a b, is_value a -> Rstar R (b**(T0;a)) (b**T0).
move => a b H.
by core_step.
Qed.

Goal forall a b x, Rstar R (a**x + b**x) ((a +s b)**x).
move => a b x.
by core_step.
Qed.

*)



(** Given [x], find the possible AC-conversions and construct a list 
    of empty trees with these terms as roots. *)
Ltac find_ACequiv r :=
    match r with
      | T0 => constr : (T0::nil)
      | (Tvar ?a) => constr : ((Tvar a)::nil)
      | (Tadd ?a ?b) =>
        let x := constr : (Tadd b a) in
        let yb := find_ACequiv b in 
        let myf := (fun x => constr : (Tadd a x)) in
        let y := maplist myf term yb in
          match a with
            | (Tadd ?e ?f) =>
              let z := constr : (Tadd e (Tadd f b)) in
                constr : (x::z::y)
            | _ => constr : (x::y)
          end
      | (Tapply ?a ?b) =>
        let yb := find_ACequiv b in 
        let myf := (fun x => constr : (Tapply a x)) in
        let y := maplist myf term yb in
        let y'a := find_ACequiv a in 
        let myg := (fun x => constr : (Tapply x b)) in
        let y' := maplist myg term y'a in
        applist y y'
      | (Tmul ?a ?b) => 
        let yb := find_ACequiv b in 
        let myg := (fun x => constr : (Tmul a x)) in
        maplist myg term yb
      | ?a => constr : (nil (A:=term))
    end.



(* Some tests 

Goal forall a b c,  Rstar R (a + (b + c)) (a + (c + b)).
Proof.
move => a b c.
try_ACequiv R core_step find_ACequiv 1.
Show Proof.
Qed.


Goal forall a b c d,  Rstar R (a + (b**(c+d))) (((b**(d+c))) + a).
Proof.
move => A b C D.
try_ACequiv R core_step find_ACequiv 1.
Show Proof.
Qed.

*)



(** Tactic that gives a potential next element using non-AC
    rewrites. The idea is that since the RW system is locally
    confluent (and normalizing), regardless of the term we choose it
    should not matter. *)

Ltac find_nonAC a :=
  let A := typeof a in
  let H := fresh in
    match a with
      | (?t + T0) => constr : (LtacYes t)
      | ((?c ** ?t) + (?d ** ?t)) => constr : (LtacYes ((c +s d)**t))
      | ((?c ** ?t) + ?t) => constr : (LtacYes ((c +s S1)**t))
      | (?t + ?t) => constr : (LtacYes ((S1 +s S1)**t))
      | (S0 ** _) => constr : (LtacYes T0)
      | (S1 ** ?t) => constr : (LtacYes t)
      | (_ ** T0) => constr : (LtacYes T0)
      | (?a ** (?b ** ?x)) => constr : (LtacYes ((Smul a b) ** x))
      | (?a ** (?s + ?t)) => constr : (LtacYes ((a ** s) + (a ** t)))
      | (Tadd ?s ?t) => 
              let x := find_nonAC t in
                match x with
                  | LtacYes ?y => constr : (LtacYes (Tadd s y))
                end
      | (Tadd ?s ?t) => 
              let x := find_nonAC s in 
                match x with
                  | LtacYes ?y => constr : (LtacYes (Tadd y t))
                end 
      | (Tapply ?s ?t) => 
              let x := find_nonAC t in 
                match x with
                  | LtacYes ?y => constr : (LtacYes (Tapply s y))
                end
      | (Tapply ?s ?t) => 
              let x := find_nonAC s in 
                match x with
                  | LtacYes ?y => constr : (LtacYes (Tapply y t))
                end 
      | (Tmul ?a ?t) => 
              let x := find_nonAC t in 
                match x with
                  | LtacYes ?y => constr : (LtacYes (Tmul a y))
                end 
      | (?u;(?s + ?t)) => 
        constr : (LtacYes ((u;s) + (u;t)))
      | ((?s + ?t);?u) =>
        constr : (LtacYes ((s;u) + (t;u)))
      | (?u;(?a ** ?t)) =>
        constr : (LtacYes (a ** (u;t)))
      | ((?a ** ?t);?u) =>
        constr : (LtacYes (a ** (t;u)))
      | (?u;T0) => constr : (LtacYes (T0))
      | (T0;?u) => constr : (LtacYes (T0))
      | ?t => match goal with
                | [ L : Rstar R t t  |- _ ] => 
                  move : L; let y := find_nonAC a in
                    move => L; constr : y(* clear trivial loops *)
                | [ L : Rstar R t ?u |- _ ] => constr : (LtacYes u)
                | [ L : R t ?u |- _ ] => constr : (LtacYes u)
              end
      | ?t => constr : (LtacNo (term:=A))
    end.


(* Some tests

Goal forall a b, 
  Rstar R a (S0 ** b) -> 
  Rstar R a T0.
move => a b H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 H).
auto.
Qed.

Goal forall a b c, 
  is_value b -> 
  is_value c ->
  Rstar R a ((S0 ** b);(c)) -> 
  Rstar R a T0.
move => a b c ? ? H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 H).
auto.
Qed.


Goal forall a u v w, 
  is_value u -> is_value v -> is_value w -> 
  Rstar R (u;(a ** (v + w))) (u;(a ** v + a ** w)).
move => a u v w H K L.
core_step.
Qed.


Goal forall a x u v w, 
  is_value u -> is_value v -> is_value w -> 
  Rstar R x (u ; (a ** (v + w))) -> 
  Rstar R x ((a ** (u;v)) + (a ** (u;w))).
move => a x u v w ? ? ? H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 H).
auto.
Qed.


Goal forall a b x u v, 
  is_value u -> is_value v ->
  Rstar R x ((a ** u); (S1**(b ** v))) ->
  Rstar R x ((Smul b a) ** (u;v)).
move => a b x u v ? ? H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 H).
auto.
Qed.


Goal forall x y, Rstar R x (T0 + (y + T0)) -> Rstar R x y.
move => x y H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 4 H).
auto.
Qed.


Goal forall a b x y, Rstar R x (a ** y + b ** y) -> Rstar R x ((a +s b) ** y).
move => a b x y H.
repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 4 H).
auto.
Qed.

*)





(** Try to add as hypothesis [H : is_value s] for all the possible [s] we can find. *)

Ltac rewrite_value :=
   match goal with
     | [ K : ?r = ?s |- _ ] =>
       try match goal with
         | [ H : is_value r |- _ ] =>
           (let L := fresh in (have L := H)); rewrite K in H; try split_value H
       end;
       try match goal with
         | [ H : is_value s |- _ ] =>
           (let L := fresh in (have L := H)); rewrite -K in H; try split_value H
       end;
       move : K; rewrite_value; move => ?
     | _ => idtac
   end.



(** make [try_smart_solve] a bit more smart *)

Lemma mul_false : forall b t, b ** t = t -> False.
Proof.
move => b.
elim.
move => H; inversion H.
move => ? ? ? ? H; inversion H.
move => ? ? K H.
inversion H.
rewrite -H1 in H2.
apply K.
done.
move => ? H; inversion H.
move => ? ? H; inversion H.
move => ? ? ? ? H; inversion H.
Qed.



Ltac try_various d1 d2 := 
  try solve [
    match goal with
      | [ H : R T0 _ |- _ ] => inversion H
      | [ H : _ ** ?s = ?s |- _ ] => 
          let A := fresh in
            (suff A : False); first (by contradiction); apply (mul_false _ _ H)
      | [ H : ?s = _ ** ?s |- _ ] => 
          symmetry in H;
          let A := fresh in
            (suff A : False); first (by contradiction); apply (mul_false _ _ H)
    end
    |
    rewrite_value; try_smart_solve R core_step find_nonAC find_ACequiv d1 d2
].


(** Define a shortcut for [go_as_far_as_possible]. *)

Ltac go_forward :=
  go_as_far_as_possible R core_step find_nonAC find_ACequiv.


(*  let H := fresh in let K := fresh in let L := fresh in
  let L1:= fresh in let L2:= fresh in let u' := fresh in
  match goal with
    | [ |- (exists u:_, Rstar R ?r _ /\ Rstar R ?s _)] => idtac "yes";
      (have H : (Rstar R r r)); first (by apply Rzero);
        (have K : (Rstar R s s)); first (by apply Rzero)
  end;
  move : H; (repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 K)); move => H;
  move : K; (repeat progress (exhaust_Rstar R core_step find_nonAC find_ACequiv 3 H)); move => K;
  move : H K;
    match goal with
      | [ |- Rstar R _ ?r' -> Rstar R _ ?s' -> _ ] => idtac r'; idtac s'; 
        move => H K;
          (suff L : (exists u, Rstar R r' u /\ Rstar R s' u)); 
          [ (elim L => u' [L1 L2]; exists u'; split; solve_Rstar_trans) | ]
    end.
*)








(** We can now show the local confluence of [R] almost automatically *)

Theorem R_local_confluence : forall r s t:term,
  (R r s) -> (R r t) -> exists u:term, (Rstar R s u) /\ (Rstar R t u).
Proof with (try_various 5 5).
elim.
(* T0 *)
intros...
(* + *)
move => s IHs t IHt u v H K.
inversion H...
rewrite <- H1 in *; clear H1.
rewrite <- H2 in *; clear H2.
rewrite -> H0 in *; clear H0.
inversion K...
rewrite <- H1 in *; clear H1.
rewrite <- H2 in *; clear H2.
clear H0.
inversion K...
inversion H2.
rewrite -H4.
rewrite -{1 2}(S_1_lunit a) -S_ldistrib...
rewrite S_add_commut...
inversion H3...
rewrite -H5 S_0_lunit...
rewrite -H5 S_add_commut...
go_forward.
rewrite -S_ldistrib...
go_forward.
rewrite (S_add_commut b a)...
inversion H3.
rewrite -H5 S_add_commut S_0_lunit...
rewrite -H5...
auto...
go_forward.
rewrite S_ldistrib...
go_forward.
rewrite (S_add_commut b a)...
auto...
rewrite -> H2 in *; clear H2.
rewrite <- H1 in *; clear H1.
inversion K...
inversion H4...
rewrite -H6 S_0_lunit...
rewrite -H6...
go_forward.
rewrite S_ldistrib S_1_lunit...
rewrite <- H2 in *; clear H2.
inversion K...
inversion H4.
rewrite <- H6 in *; clear H6.
go_forward.
rewrite S_ldistrib S_1_lunit...
rewrite <- H1 in *; clear H1.
inversion K...
try_various 0 8.
rewrite <- H2 in *.
inversion K...
inversion K...
rewrite -H4...
rewrite <- H5 in *.
inversion H3...
rewrite -H8 S_0_lunit...
rewrite -H8 S_add_commut...
go_forward.
rewrite S_ldistrib...
go_forward.
rewrite S_add_commut...
rewrite <- H5 in *; clear H5.
inversion H3...
rewrite -H7 S_0_lunit...
rewrite -H5 H6 -H7...
rewrite -H6 -H8...
rewrite -H6 -H8.
go_forward.
rewrite S_ldistrib S_1_lunit...
rewrite -H6 -H8...
rewrite -H6...
rewrite H6 in H3...
rewrite <- H5 in *; clear H5.
inversion H3...
try_various 0 8.
elim (IHs _ _ H3 H7) => x [L1 L2]...
inversion K...
rewrite -H6 in H3; inversion H3.
rewrite <- H6 in *; clear H6.
inversion H3...
rewrite -H7 S_add_commut S_0_lunit...
rewrite -H7...
go_forward.
rewrite S_ldistrib...
go_forward; rewrite S_add_commut...
rewrite <- H6 in *; clear H6.
inversion H3...
try_various 0 8.
elim (IHt _ _ H3 H7) => x [L1 L2]...
(* ** *)
move => a r IHr s t H K.
inversion H...
rewrite <- H1 in *.
inversion K...
rewrite S_0_lelim...
rewrite <- H1 in *. 
inversion K...
symmetry in H4.
elim (S_0_1_dec H4).
rewrite -H0 -H3...
rewrite -H0 -H5...
rewrite S_1_lunit.
rewrite H5 H0...
rewrite -H0 -H5...
rewrite -H0...
rewrite <- H2 in *.
inversion K...
rewrite <- H2 in *.
inversion K...
rewrite <- H3 in *.
rewrite <- H4 in *.
rewrite  S_0_lelim...
rewrite <- H4 in *.
rewrite S_1_lunit...
inversion H6...
rewrite -H8 S_mul_commut S_0_lelim...
rewrite -H8 S_mul_commut S_1_lunit...
go_forward.
rewrite S_mul_assoc...
rewrite <- H2 in *.
inversion K...
rewrite -H4...
rewrite -H4...
inversion H6...
go_forward.
rewrite S_rdistrib...
go_forward.
rewrite S_rdistrib (S_mul_commut a S1) S_1_lunit...
go_forward.
rewrite S_rdistrib S_mul_commut S_1_lunit...
inversion K...
rewrite -H5...
rewrite <- H5 in *; clear H5.
rewrite <- H4 in *; clear H4.
auto...
rewrite <- H6 in *; clear H6.
rewrite <- H4 in *; clear H4.
rewrite <- H1 in *; clear H1.
inversion H...
rewrite <- H6 in *; clear H6.
rewrite <- H4 in *; clear H4.
rewrite <- H1 in *; clear H1.
inversion H3...
rewrite -H4 -S_mul_commut S_0_lelim...
rewrite -H4 S_mul_commut S_1_lunit...
go_forward.
rewrite S_mul_assoc...
rewrite <- H6 in *.
inversion H3...
go_forward.
rewrite S_rdistrib...
go_forward.
rewrite S_rdistrib (S_mul_commut a S1) S_1_lunit...
go_forward.
rewrite S_rdistrib S_mul_commut S_1_lunit...
elim (IHr _ _ H3 H7) => x [L1 L2]...
(* @n *)
move => ? ? ? H ?.
inversion H.
(* \ t *)
move => ? ? ? ? ? H; inversion H.
(* s ; t *)
move => r IHr s IHs t u H K.
inversion H.
inversion K...
elim (IHr _ _ H3 H7) => x [L1 L2]...
rewrite <- H4 in *; clear H4.
show_is_value M (r0 + s1).
inversion H3...
rewrite <- H4 in *; clear H4.
show_is_value M (a ** r0).
inversion H3...
rewrite -H9...
rewrite -H9...
rewrite <- H4 in H3; inversion H3.
rewrite <- H3 in *; clear H3.
clear H0 H1.
inversion K...
elim (IHs _ _ H4 H6) => x [L1 L2]...
rewrite <- H1 in *; clear H1.
inversion H4...
rewrite <- H1 in *; clear H1.
inversion H4...
rewrite -H7...
rewrite -H7...
rewrite <- H3 in H4; inversion H4.
rewrite <- H0 in *; clear H0 H1.
inversion K...
inversion H7...
rewrite <- H1 in H5; (try_various 5 8).
rewrite <- H0 in *; clear H0.
inversion K...
inversion H7...
rewrite -H9...
rewrite -H9...
go_forward.
rewrite S_mul_commut...
rewrite <- H0 in K; inversion K...
rewrite <- H1 in *; clear H1 H0 H4.
inversion K...
inversion H7...
rewrite <- H0 in H5; (try_various 5 8).
rewrite <- H1 in *; clear H1 H3 H0.
inversion K...
inversion H6...
rewrite -H8...
rewrite -H8...
go_forward.
rewrite S_mul_commut...
rewrite <- H2 in *; clear H2.
inversion K...
Qed.

End Llin.

